@@ -13,17 +13,24 @@
 --   Support for SSL is now obsoleted.
 
 module Network.Wai.Handler.WarpTLS (
+    -- * Runner
+      runTLS
+    , runTLSSocket
     -- * Settings
-      TLSSettings
+    , TLSSettings
     , defaultTlsSettings
     -- * Smart constructors
+    -- ** From files
     , tlsSettings
-    , tlsSettingsMemory
     , tlsSettingsChain
+    -- ** From memory
+    , tlsSettingsMemory
     , tlsSettingsChainMemory
+    -- ** From references
+    , tlsSettingsRef
+    , tlsSettingsChainRef
+    , CertSettings
     -- * Accessors
-    , certFile
-    , keyFile
     , tlsCredentials
     , tlsLogging
     , tlsAllowedVersions
@@ -35,18 +42,19 @@ module Network.Wai.Handler.WarpTLS (
     , tlsSessionManager
     , onInsecure
     , OnInsecure (..)
-    -- * Runner
-    , runTLS
-    , runTLSSocket
     -- * Exception
     , WarpTLSException (..)
+    -- * DH parameters (re-exports)
+    --
+    -- | This custom DH parameters are not necessary anymore because
+    --   pre-defined DH parameters are supported in the TLS package.
     , DH.Params
     , DH.generateParams
     ) where
 
 import Control.Applicative ((<|>))
-import Control.Exception (Exception, throwIO, bracket, finally, handle, fromException, try, IOException, onException, SomeException(..), handleJust)
-import qualified Control.Exception as E
+import UnliftIO.Exception (Exception, throwIO, bracket, finally, handle, fromException, try, IOException, onException, SomeException(..), handleJust)
+import qualified UnliftIO.Exception as E
 import Control.Monad (void, guard)
 import qualified Data.ByteString as S
 import qualified Data.ByteString.Lazy as L
@@ -67,120 +75,19 @@ import qualified Network.TLS.SessionManager as SM
 import Network.Wai (Application)
 import Network.Wai.Handler.Warp
 import Network.Wai.Handler.Warp.Internal
+import Network.Wai.Handler.WarpTLS.Internal(CertSettings(..), TLSSettings(..), OnInsecure(..))
 import System.IO.Error (isEOFError, ioeGetErrorType)
 
-----------------------------------------------------------------
+-- | The default 'CertSettings'.
+defaultCertSettings :: CertSettings
+defaultCertSettings = CertFromFile "certificate.pem" [] "key.pem"
 
--- | Settings for WarpTLS.
-data TLSSettings = TLSSettings {
-    certFile :: FilePath
-    -- ^ File containing the certificate.
-  , chainCertFiles :: [FilePath]
-    -- ^ Files containing chain certificates.
-  , keyFile :: FilePath
-    -- ^ File containing the key
-  , certMemory :: Maybe S.ByteString
-  , chainCertsMemory :: [S.ByteString]
-  , keyMemory :: Maybe S.ByteString
-  , onInsecure :: OnInsecure
-    -- ^ Do we allow insecure connections with this server as well?
-    --
-    -- >>> onInsecure defaultTlsSettings
-    -- DenyInsecure "This server only accepts secure HTTPS connections."
-    --
-    -- Since 1.4.0
-  , tlsLogging :: TLS.Logging
-    -- ^ The level of logging to turn on.
-    --
-    -- Default: 'TLS.defaultLogging'.
-    --
-    -- Since 1.4.0
-  , tlsAllowedVersions :: [TLS.Version]
-#if MIN_VERSION_tls(1,5,0)
-    -- ^ The TLS versions this server accepts.
-    --
-    -- >>> tlsAllowedVersions defaultTlsSettings
-    -- [TLS13,TLS12,TLS11,TLS10]
-    --
-    -- Since 1.4.2
-#else
-    -- ^ The TLS versions this server accepts.
-    --
-    -- >>> tlsAllowedVersions defaultTlsSettings
-    -- [TLS12,TLS11,TLS10]
-    --
-    -- Since 1.4.2
-#endif
-  , tlsCiphers :: [TLS.Cipher]
-#if MIN_VERSION_tls(1,5,0)
-    -- ^ The TLS ciphers this server accepts.
-    --
-    -- >>> tlsCiphers defaultTlsSettings
-    -- [ECDHE-ECDSA-AES256GCM-SHA384,ECDHE-ECDSA-AES128GCM-SHA256,ECDHE-RSA-AES256GCM-SHA384,ECDHE-RSA-AES128GCM-SHA256,DHE-RSA-AES256GCM-SHA384,DHE-RSA-AES128GCM-SHA256,ECDHE-ECDSA-AES256CBC-SHA384,ECDHE-RSA-AES256CBC-SHA384,DHE-RSA-AES256-SHA256,ECDHE-ECDSA-AES256CBC-SHA,ECDHE-RSA-AES256CBC-SHA,DHE-RSA-AES256-SHA1,RSA-AES256GCM-SHA384,RSA-AES256-SHA256,RSA-AES256-SHA1,AES128GCM-SHA256,AES256GCM-SHA384]
-    --
-    -- Since 1.4.2
-#else
-    -- ^ The TLS ciphers this server accepts.
-    --
-    -- >>> tlsCiphers defaultTlsSettings
-    -- [ECDHE-ECDSA-AES256GCM-SHA384,ECDHE-ECDSA-AES128GCM-SHA256,ECDHE-RSA-AES256GCM-SHA384,ECDHE-RSA-AES128GCM-SHA256,DHE-RSA-AES256GCM-SHA384,DHE-RSA-AES128GCM-SHA256,ECDHE-ECDSA-AES256CBC-SHA384,ECDHE-RSA-AES256CBC-SHA384,DHE-RSA-AES256-SHA256,ECDHE-ECDSA-AES256CBC-SHA,ECDHE-RSA-AES256CBC-SHA,DHE-RSA-AES256-SHA1,RSA-AES256GCM-SHA384,RSA-AES256-SHA256,RSA-AES256-SHA1]
-    --
-    -- Since 1.4.2
-#endif
-  , tlsWantClientCert :: Bool
-    -- ^ Whether or not to demand a certificate from the client.  If this
-    -- is set to True, you must handle received certificates in a server hook
-    -- or all connections will fail.
-    --
-    -- >>> tlsWantClientCert defaultTlsSettings
-    -- False
-    --
-    -- Since 3.0.2
-  , tlsServerHooks :: TLS.ServerHooks
-    -- ^ The server-side hooks called by the tls package, including actions
-    -- to take when a client certificate is received.  See the "Network.TLS"
-    -- module for details.
-    --
-    -- Default: def
-    --
-    -- Since 3.0.2
-  , tlsServerDHEParams :: Maybe DH.Params
-    -- ^ Configuration for ServerDHEParams
-    -- more function lives in `cryptonite` package
-    --
-    -- Default: Nothing
-    --
-    -- Since 3.2.2
-  , tlsSessionManagerConfig :: Maybe SM.Config
-    -- ^ Configuration for in-memory TLS session manager.
-    -- If Nothing, 'TLS.noSessionManager' is used.
-    -- Otherwise, an in-memory TLS session manager is created
-    -- according to 'Config'.
-    --
-    -- Default: Nothing
-    --
-    -- Since 3.2.4
-  , tlsCredentials :: Maybe TLS.Credentials
-    -- ^ Specifying 'TLS.Credentials' directly.  If this value is
-    --   specified, other fields such as 'certFile' are ignored.
-    --
-    --   Since 3.2.12
-  , tlsSessionManager :: Maybe TLS.SessionManager
-    -- ^ Specifying 'TLS.SessionManager' directly. If this value is
-    --   specified, 'tlsSessionManagerConfig' is ignored.
-    --
-    --   Since 3.2.12
-  }
+----------------------------------------------------------------
 
 -- | Default 'TLSSettings'. Use this to create 'TLSSettings' with the field record name (aka accessors).
 defaultTlsSettings :: TLSSettings
 defaultTlsSettings = TLSSettings {
-    certFile = "certificate.pem"
-  , chainCertFiles = []
-  , keyFile = "key.pem"
-  , certMemory = Nothing
-  , chainCertsMemory = []
-  , keyMemory = Nothing
+    certSettings = defaultCertSettings
   , onInsecure = DenyInsecure "This server only accepts secure HTTPS connections."
   , tlsLogging = def
 #if MIN_VERSION_tls(1,5,0)
@@ -203,20 +110,12 @@ ciphers = TLSExtra.ciphersuite_strong
 
 ----------------------------------------------------------------
 
--- | An action when a plain HTTP comes to HTTP over TLS/SSL port.
-data OnInsecure = DenyInsecure L.ByteString
-                | AllowInsecure
-                deriving (Show)
-
-----------------------------------------------------------------
-
 -- | A smart constructor for 'TLSSettings' based on 'defaultTlsSettings'.
 tlsSettings :: FilePath -- ^ Certificate file
             -> FilePath -- ^ Key file
             -> TLSSettings
 tlsSettings cert key = defaultTlsSettings {
-    certFile = cert
-  , keyFile = key
+    certSettings = CertFromFile cert [] key
   }
 
 -- | A smart constructor for 'TLSSettings' that allows specifying
@@ -229,9 +128,7 @@ tlsSettingsChain
             -> FilePath -- ^ Key file
             -> TLSSettings
 tlsSettingsChain cert chainCerts key = defaultTlsSettings {
-    certFile = cert
-  , chainCertFiles = chainCerts
-  , keyFile = key
+    certSettings = CertFromFile cert chainCerts key
   }
 
 -- | A smart constructor for 'TLSSettings', but uses in-memory representations
@@ -242,10 +139,9 @@ tlsSettingsMemory
     :: S.ByteString -- ^ Certificate bytes
     -> S.ByteString -- ^ Key bytes
     -> TLSSettings
-tlsSettingsMemory cert key = defaultTlsSettings
-    { certMemory = Just cert
-    , keyMemory = Just key
-    }
+tlsSettingsMemory cert key = defaultTlsSettings { 
+    certSettings = CertFromMemory cert [] key
+  }
 
 -- | A smart constructor for 'TLSSettings', but uses in-memory representations
 -- of the certificate and key based on 'defaultTlsSettings'.
@@ -256,11 +152,34 @@ tlsSettingsChainMemory
     -> [S.ByteString] -- ^ Chain certificate bytes
     -> S.ByteString -- ^ Key bytes
     -> TLSSettings
-tlsSettingsChainMemory cert chainCerts key = defaultTlsSettings
-    { certMemory = Just cert
-    , chainCertsMemory = chainCerts
-    , keyMemory = Just key
-    }
+tlsSettingsChainMemory cert chainCerts key = defaultTlsSettings { 
+    certSettings = CertFromMemory cert chainCerts key
+  }
+
+-- | A smart constructor for 'TLSSettings', but uses references to in-memory
+-- representations of the certificate and key based on 'defaultTlsSettings'.
+--
+-- @since 3.3.0
+tlsSettingsRef 
+    :: I.IORef S.ByteString -- ^ Reference to certificate bytes
+    -> I.IORef (S.ByteString) -- ^ Reference to key bytes 
+    -> TLSSettings 
+tlsSettingsRef cert key = defaultTlsSettings { 
+    certSettings = CertFromRef cert [] key
+  }
+
+-- | A smart constructor for 'TLSSettings', but uses references to in-memory
+-- representations of the certificate and key based on 'defaultTlsSettings'.
+--
+-- @since 3.3.0
+tlsSettingsChainRef 
+    :: I.IORef S.ByteString -- ^ Reference to certificate bytes
+    -> [I.IORef S.ByteString] -- ^ Reference to chain certificate bytes
+    -> I.IORef (S.ByteString) -- ^ Reference to key bytes 
+    -> TLSSettings 
+tlsSettingsChainRef cert chainCerts key = defaultTlsSettings { 
+    certSettings = CertFromRef cert chainCerts key
+  }
 
 ----------------------------------------------------------------
 
@@ -276,15 +195,19 @@ runTLS tset set app = withSocketsDo $
 
 loadCredentials :: TLSSettings -> IO TLS.Credentials
 loadCredentials TLSSettings{ tlsCredentials = Just creds } = return creds
-loadCredentials TLSSettings{..} = case (certMemory, keyMemory) of
-    (Nothing, Nothing) -> do
-        cred <- either error id <$> TLS.credentialLoadX509Chain certFile chainCertFiles keyFile
-        return $ TLS.Credentials [cred]
-    (mcert, mkey) -> do
-        cert <- maybe (S.readFile certFile) return mcert
-        key <- maybe (S.readFile keyFile) return mkey
-        cred <- either error return $ TLS.credentialLoadX509ChainFromMemory cert chainCertsMemory key
-        return $ TLS.Credentials [cred]
+loadCredentials TLSSettings{..} = case certSettings of 
+  CertFromFile cert chainFiles key -> do
+    cred <- either error id <$> TLS.credentialLoadX509Chain cert chainFiles key
+    return $ TLS.Credentials [cred]
+  CertFromRef certRef chainCertsRef keyRef -> do 
+    cert <- I.readIORef certRef
+    chainCerts <- mapM I.readIORef chainCertsRef
+    key <- I.readIORef keyRef
+    cred <- either error return $ TLS.credentialLoadX509ChainFromMemory cert chainCerts key
+    return $ TLS.Credentials [cred]
+  CertFromMemory certMemory chainCertsMemory keyMemory -> do
+    cred <- either error return $ TLS.credentialLoadX509ChainFromMemory certMemory chainCertsMemory keyMemory
+    return $ TLS.Credentials [cred]
 
 getSessionManager :: TLSSettings -> IO TLS.SessionManager
 getSessionManager TLSSettings{ tlsSessionManager = Just mgr } = return mgr
@@ -346,7 +269,7 @@ alpn xs
 ----------------------------------------------------------------
 
 getter :: TLS.TLSParams params => TLSSettings -> Settings -> Socket -> params -> IO (IO (Connection, Transport), SockAddr)
-getter tlsset@TLSSettings{..} set sock params = do
+getter tlsset set sock params = do
 #if WINDOWS
     (s, sa) <- windowsThreadBlockHack $ accept sock
 #else
@@ -356,14 +279,12 @@ getter tlsset@TLSSettings{..} set sock params = do
     return (mkConn tlsset set s params, sa)
 
 mkConn :: TLS.TLSParams params => TLSSettings -> Settings -> Socket -> params -> IO (Connection, Transport)
-mkConn tlsset set s params = switch `onException` close s
+mkConn tlsset set s params = (safeRecv s 4096 >>= switch) `onException` close s
   where
-    switch = do
-        firstBS <- safeRecv s 4096
-        if not (S.null firstBS) && S.head firstBS == 0x16 then
-            httpOverTls tlsset set s firstBS params
-          else
-            plainHTTP tlsset set s firstBS
+    switch firstBS
+        | S.null firstBS = close s >> throwIO ClientClosedConnectionPrematurely
+        | S.head firstBS == 0x16 = httpOverTls tlsset set s firstBS params
+        | otherwise = plainHTTP tlsset set s firstBS
 
 ----------------------------------------------------------------
 
@@ -553,6 +474,8 @@ recvPlain ref fallback = do
 
 ----------------------------------------------------------------
 
-data WarpTLSException = InsecureConnectionDenied
+data WarpTLSException
+    = InsecureConnectionDenied
+    | ClientClosedConnectionPrematurely
     deriving (Show, Typeable)
 instance Exception WarpTLSException
